/* VM profiling support stuff

  Copyright (C) 2001 Free Software Foundation, Inc.

  This file is part of Gforth.

  Gforth is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*/

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "forth.h"


/* data structure: simple hash table with external chaining */

#define HASH_SIZE (1<<20)

typedef struct block_count {
  struct block_count *next; /* next in hash table */
  struct block_count *fallthrough; /* the block that this one falls
                                       through to without SUPER_END */
  Xt *ip;
  long long count;
  char **insts;
  size_t ninsts;
} block_count;

block_count *blocks[HASH_SIZE];

#define hash(p) ((((Cell)(p))/sizeof(Xt))&(HASH_SIZE-1))

block_count *block_lookup(Xt *ip)
{
  block_count *b = blocks[hash(ip)];

  while (b!=NULL && b->ip!=ip)
    b = b->next;
  return b;
}

/* looks up present elements, inserts absent elements */
block_count *block_insert(Xt *ip)
{ 
  block_count *b = block_lookup(ip);
  block_count *new;

  if (b != NULL)
    return b;
  new = (block_count *)malloc(sizeof(block_count));
  new->next = blocks[hash(ip)];
  new->fallthrough = NULL;
  new->ip = ip;
  new->count = 0LL;
  new->insts = malloc(0);
  assert(new->insts != NULL);
  new->ninsts = 0;
  blocks[hash(ip)] = new;
  return new;
}

void add_inst(block_count *b, char *inst)
{
  b->insts = realloc(b->insts, (b->ninsts+1) * sizeof(char *));
  b->insts[b->ninsts++] = inst;
}

void vm_count_block(Xt *ip)
{
  block_insert(ip)->count++;
}

char *prim_name[] = {
  "docol",
  "docon",
  "dovar",
  "douser",
  "dodefer",
  "dofield",
  "dodoes",
  "doesjump",
#include "prim_name.i"
  NULL
};

int lookup_prim(char *name)
{
  int i;

  for (i=0; prim_name[i] != NULL; i++)
    if (strcmp(prim_name[i],name)==0)
      return i;
  fprintf(stderr, "Primitive not found: %s\n", name);
  return i;
}

#define MAXPRIMS 100000
Cell orig_prim[MAXPRIMS]; /* orig_prim[i] is the original prim number
                             of the copy i */
Cell ends_with_exec[MAXPRIMS]; /* 1 if the primitive ends with execute etc. */

Cell doer_replacements[] = {333/*call*/, 9/*lit*/, 9/*lit*/, 334/*useraddr*/,
			    4/*dodefer*/, 5/*dofield*/, 6/*dodoes*/, 0 , 0};
Cell branch_alt_jumps[MAXPRIMS];

typedef struct {
  int firstcopy;
  int lastcopy;
  int nextcopy;
} replication_t;

replication_t replication[MAXPRIMS];

void read_replication_file(char *filename)
{
  FILE *file=fopen(filename,"r");
  char buf[128];
  Cell prim;

  while (fgets(buf, 128, file)) {
    int buflen=strlen(buf);
    char *endname = strchr(buf, ' ');
    char *startname;
    int copies;

    if (buf[buflen-1]=='\n')
      buf[buflen-1]='\0';
    *endname = 0;
    startname = endname+1;
    copies = atoi(buf);
    if (copies<1) {
      fprintf(stderr, "Copies must be >0: %s %s\n", buf, startname);
      continue;
    }
    endname = strchr(startname, ' ');
    if (endname)
      *endname = 0;
    prim = lookup_prim(startname);
    while (endname) {
      startname = endname+1;
      endname = strchr(startname, ' ');
      if (endname)
	*endname = 0;
      prim = insert_combo(prim,lookup_prim(startname));
      assert(prim != -1);
    }
    replication[prim].firstcopy= max_prim;
    replication[prim].lastcopy = max_prim+copies-1;
    replication[prim].nextcopy = max_prim;
    max_prim += copies;
  }
}

void note_prim(int prim, Address addr)
{
  static int last_prim = -1;
  static Cell *last_addr = 0;

  if (prim>=0 && prim<MAXORIGPRIM) {
    Cell newprim;
    Cell combo;
    Cell lastpart = prim;

    if (prim <= DODOES)
      prim = doer_replacements[prim];
    if (last_addr!=NULL && block_lookup((Xt *)addr)==NULL &&
	(combo=lookup_combo(last_prim,prim))!=-1) {
      *last_addr=-1;
      prim = combo;
    }
    last_prim = prim;
    last_addr = (Cell *)(addr+shadow_offset);
    if (replication[prim].firstcopy != 0) {
      if (random_replication) {
	replication_t *r=&replication[prim];
	newprim = r->firstcopy + (random() % (r->lastcopy - r->firstcopy + 1));
      } else {
      newprim = replication[prim].nextcopy;
      replication[prim].nextcopy++;
      if (replication[prim].nextcopy > replication[prim].lastcopy)
	replication[prim].nextcopy = replication[prim].firstcopy;
      }
    } else if (use_replication)
      newprim = max_prim++;
    else
      newprim = prim;
    *last_addr = newprim;
    orig_prim[newprim] = prim;
    ends_with_exec[newprim] = (lastpart==4 || lastpart==10 || lastpart==11);
    if (lastpart == 0 || lastpart == 4 || lastpart == 6 || 
	(lastpart>=10 && lastpart < 36 && lastpart != 28) ||
	lastpart == 125 || lastpart == 333) { /* bb ends */
      last_prim = -1;
      last_addr = 0;
    }
  }
}

void compile_note(Xt xt, Address addr)
     /* xt is compiled at addr; assumes indirect threaded code */
{
  int i;
  unsigned prim = xt-vm_prims;

  if (start_bb)
    block_insert((Xt *)addr);
  start_bb=1; /* cleared by peephole_opt */
  if (prim > MAXORIGPRIM) { /* not a primitive */
    Label ca = *xt;

    for (i=0; ca!=vm_prims[i]; i++)
      assert(i<=DODOES);
    prim = i;
  }
  note_prim(prim, addr);
}

#ifdef DIRECT_THREADED
#define VM_IS_INST(inst, n) ((inst) == vm_prims[(n)+DOESJUMP+1])
#else
#define VM_IS_INST(inst, n) ((inst) == &(vm_prims[(n)+DOESJUMP+1]))
#endif

void postprocess_block(block_count *b)
{
  Xt *ip = b->ip;
  block_count *next_block;

  do {
#include "profile.i"
    /* else */
    {
      add_inst(b,"unknown");
      ip++;
    }
  _endif_:
    next_block = block_lookup(ip);
  } while (next_block == NULL);
  /* we fell through, so set fallthrough and update the count */
  b->fallthrough = next_block;
  /* also update the counts of all following fallthrough blocks that
     have already been processed */
  while (next_block != NULL) {
    next_block->count += b->count;
    next_block = next_block->fallthrough;
  }
}

/* Deal with block entry by falling through from non-SUPER_END
   instructions.  And fill the insts and ninsts fields. */
void postprocess(void)
{
  size_t i;

  for (i=0; i<HASH_SIZE; i++) {
    block_count *b = blocks[i];
    for (; b!=0; b = b->next)
      postprocess_block(b);
   }
}

#if 0
void print_block(FILE *file, block_count *b)
{
  size_t i;

  fprintf(file,"%14lld\t",b->count);
  for (i=0; i<b->ninsts; i++)
    fprintf(file, "%s ", b->insts[i]);
  putc('\n', file);
}
#endif

void print_block(FILE *file, block_count *b)
{
  size_t i,j,k;

  for (i=2; i<12; i++)
    for (j=0; i+j<=b->ninsts; j++) {
      fprintf(file,"%14lld\t",b->count);
      for (k=j; k<i+j; k++)
	fprintf(file, "%s ", b->insts[k]);
      putc('\n', file);
    }
}

/* btb tracing */

typedef struct {
  int last_prim_id;
  unsigned execs;
  unsigned mispreds;
} btb_count_t;

btb_count_t btb_counts[MAXPRIMS+1];

int recent_branches[MAXPRIMS];  /* recent_branches[i] is the i-th most
                                   recent branch */
/* int recency_ranking[MAXPRIMS][MAXPRIMS+1];*/
/* recency_ranking[i][j] contains the number of times that branch i
   was on the j-th rank when it was executed */

Cell recency_btb_count[MAXPRIMS+1];    /* btb hit recency counts */
Cell recency_icache_count[MAXPRIMS+1]; /* all recency counts */

/* double conflict_hits[20]; *//* conflict_hits[i] is the number of
                             expected hits in a BTB of size i, taking
                             conflict misses into consideration, but
                             excluding lucky hits */

void init_stuff()
{
  int i;
  
  for (i=0; i<MAXPRIMS+1; i++) {
    btb_counts[i].last_prim_id = -1;
  }
  for (i=0; i<MAXORIGPRIM; i++)
    orig_prim[i] = i;
}

int compute_recency(int branch)
     /* In a sequence of calls to compute_recency, this returns how
        many different other branches were used since the last time
        compute_recency was called for "branch";
	
	This is implemented through linear seach in a move-to-front array.

	On the first lookup for a given branch we get a capacity miss
	and compute_recency returns MAX_PRIMS (except for noop (prim 0)).
     */
{
  int i;
  int temp1, temp2;

  for (i=0, temp1=recent_branches[0]; branch!=temp1 && i<MAXPRIMS; i++) {
    temp2 = recent_branches[i+1];
    recent_branches[i+1] = temp1;
    temp1 = temp2;
  }
  recent_branches[0] = branch;
  return i;
}

Cell dispatches=0;

/* this version does not use ca as branch target; instead, it uses the
   prim_id, and always is one step behind (because we don't know the
   next id in advance) */

void trace_btb_ca(int i, char *name, Label ca, Xt *ip1)
     /* branch number within prim i, prim-name, code address, ip+1 */
{
  int prim_id = *(Cell *)(((Address)(ip1-1))+shadow_offset);
  btb_count_t *b;
  int recency;
  static int last_branch_num=MAXPRIMS-1;
  static Label last_ca;
  
  dispatches++;
  if (prim_id != -1) {
    assert(prim_id != 0 || strcmp(name,"docol")==0);
    if (ends_with_exec[prim_id]) { /* !! or ends_with_exec[orig_prim[prim_id]]? */
      /* execute, perform, dodefer */
      int j;
      for (j=0; vm_prims[j]!=last_ca; j++)
	assert(j==7 || vm_prims[j]!=NULL);
      if (!ends_with_exec[j]) /* only replace prim_id if we are
                                 tracing the executed xt; checking for
                                 ends_with_exec is not totally
                                 correct, but close enough */
	prim_id = j;
    }
    b = &btb_counts[last_branch_num];
    recency = compute_recency(last_branch_num);
    recency_icache_count[recency]++;
    b->execs++;
    if (prim_id != b->last_prim_id)
      b->mispreds++;
    else {
      if (recency == MAXPRIMS)
	fprintf(stderr, "prim_id=%d, last_branch_num=%d, b->last_prim_id=%d\n",
		prim_id, last_branch_num, b->last_prim_id);
      /* assert(recency != MAXPRIMS);*/
      recency_btb_count[recency]++;
    }
    b->last_prim_id = prim_id;
    last_branch_num = prim_id;
    if (i & use_branch_alts) { /* alternate jump for a conditional branch */
      assert(i==1);
      if (branch_alt_jumps[prim_id] == 0) {
	orig_prim[max_prim] = orig_prim[prim_id];
	branch_alt_jumps[prim_id] = max_prim++;
      }
      last_branch_num = branch_alt_jumps[prim_id];
    }
  }
  last_ca = ca;
}

void fprint_name(FILE *file, int prim)
{
  if (prim<MAXORIGPRIM)
    fprintf(file, "%s", prim_name[prim]);
  else {
    int i;
    
    for (i=0; i<max_combo; i++) {
      Combination *c = &peephole_table[i];

      if (prim == c->combination_prim) {
	fprint_name(file, c->prefix);
	fprintf(file, " ");
	fprint_name(file, c->lastprim);
	return;
      }
    }
    assert(0);
  }
}

void print_trace_btb(FILE *file)
{
  int i, j, logsize, btb_sum, icache_sum;
  int infinite_hits=0; /* the number of btb hits in an infinite-sized BTB */
  int infinite_misses=0;
  int total_execs=0;   /* number of executed NEXTs */
  double btb_hits, icache_hits, r;

  for (i=0; i<MAXPRIMS; i++) {
    btb_count_t *b = &btb_counts[i];
    int btb_hits = b->execs - b->mispreds;

    infinite_hits += btb_hits;
    infinite_misses += b->mispreds;
    total_execs += b->execs;
    if (b->execs > 0) {
      fprintf(file, "%15u %15u %5d ",
	      b->execs, b->mispreds, i);
      fprint_name(file, orig_prim[i]);
      fprintf(file, "\n");
#if 0
      /* checking code */
      for (j=0, sum=0; j<MAXPRIMS+1; j++)
	sum+=recency_ranking[i][j];
      /* fprintf(file,"sum: %d; execs: %d\n", sum, b->execs); */
      assert(sum == btb_hits);
      assert(recency_ranking[i][MAXPRIMS] == 0);
      for (logsize=1; logsize<20; logsize++) {
	double size=(double)(1<<logsize);
	/* numerical condition: all summands have the same sign, so
           the error should be at most 2*MAXPRIMS*ulp or so */
	for (hits=0.0, j=0, r=1.0; j<MAXPRIMS; j++, r *= (size-1.0)/size)
	  hits += r * recency_ranking[i][j];
	/* fprintf(file, " %5f", hits/(double)btb_hits); */
	conflict_hits[logsize] += hits;
      }
      /* for (j=0; j<12; j++)
	 fprintf(file, " %d", recency_ranking[i][j]);
      fprintf(file,"\n"); */
#endif
    }
  }
  fprintf(file, "%15u %15u %s\n", total_execs, infinite_misses, "Total");
  fprintf(file, "%15u %15s %s\n", dispatches, "", "No-Super-Total");
  for (j=0, btb_sum=0, icache_sum=0; j<MAXPRIMS+1; j++) {
    btb_sum += recency_btb_count[j];
    icache_sum += recency_icache_count[j];
  }
  assert(recency_btb_count[MAXPRIMS] == 0);  /* no compulsory misses */
  assert(infinite_hits == btb_sum);
  assert(total_execs == icache_sum);
  for (logsize=1; logsize<20; logsize++) {
    double size=(double)(1<<logsize);
    double btb_misses;
    /* numerical condition: all summands have the same sign, so
       the error should be at most 2*MAXPRIMS*ulp or so */
    for (btb_hits=0.0, icache_hits=0.0, j=0, r=1.0;
	 j<MAXPRIMS;
	 j++, r *= (size-1.0)/size) {
      btb_hits += r * recency_btb_count[j];
      icache_hits += r * recency_icache_count[j];
    }
    btb_misses = ((double)total_execs)-btb_hits;
    fprintf(file, "%7d: %8f %8f %8f %8f\n", 1<<logsize,
	    btb_misses/(double)dispatches,
	    btb_misses/(double)total_execs,
	    btb_hits/(double)infinite_hits,
	    icache_hits/(double)total_execs);
  }
  fprintf(file,"    inf: %8f %8f %8f %8f\n",
	  (double)infinite_misses/(double)dispatches,
	  (double)infinite_misses/(double)total_execs, 
	  1.0,
	  (double)(total_execs-recency_icache_count[MAXPRIMS])/(double)total_execs);
}

void vm_print_profile(FILE *file)
{
  size_t i;

  postprocess();
  for (i=0; i<HASH_SIZE; i++) {
    block_count *b = blocks[i];
    for (; b!=0; b = b->next)
      print_block(file, b);
   }
}
